# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import argparse
from collections.abc import Sequence

try:
    import argcomplete
except ImportError:
    HAS_ARGCOMPLETE = False
else:
    HAS_ARGCOMPLETE = True

from fedrq.cli.base import CheckConfig, Command
from fedrq.cli.commands.pkgs import Pkgs
from fedrq.cli.commands.repolist import Repolist
from fedrq.cli.commands.subpkgs import Subpkgs
from fedrq.cli.commands.whatrequires import (
    WhatCommand,
    Whatenhances,
    Whatobsoletes,
    Whatrecommends,
    Whatrequires,
    WhatrequiresSrc,
    Whatsuggests,
    Whatsupplements,
)

__all__ = (
    "Command",
    "Pkgs",
    "Repolist",
    "Subpkgs",
    "WhatCommand",
    "Whatenhances",
    "Whatobsoletes",
    "Whatrecommends",
    "Whatrequires",
    "WhatrequiresSrc",
    "Whatsuggests",
    "Whatsupplements",
)


def main(argv: Sequence | None = None, **kwargs) -> None:
    parser = argparse.ArgumentParser(
        description="fedrq is a tool for querying the Fedora and EPEL repositories.",
        **kwargs,
    )
    subparsers = parser.add_subparsers(
        title="Subcommands", dest="action", required=True
    )
    for name, cls in COMMANDS.items():
        cls.make_parser(subparsers.add_parser, name=name, add_help=True)
    if HAS_ARGCOMPLETE:
        argcomplete.autocomplete(parser)
    args = parser.parse_args(argv)
    return COMMANDS[args.action](args).run()


COMMANDS: dict[str, type[Command]] = {
    "check-config": CheckConfig,
    "pkgs": Pkgs,
    "subpkgs": Subpkgs,
    "repolist": Repolist,
    "whatenhances": Whatenhances,
    "whatobsoletes": Whatobsoletes,
    "whatrecommends": Whatrecommends,
    "whatrequires": Whatrequires,
    "wr": Whatrequires,
    "whatrequires-src": WhatrequiresSrc,
    "wrsrc": WhatrequiresSrc,
    "whatsuggests": Whatsuggests,
    "whatsupplements": Whatsupplements,
}



# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import argparse
import sys
from collections import abc as cabc

from pydantic import ValidationError

from fedrq.cli.base import Command


class Repolist(Command):
    """
    Display list of enabled repositories
    """

    def __init__(self, args: argparse.Namespace) -> None:
        self.args = args
        self._v_errors: list[str] = []

        self.v_logging()
        try:
            self.config = self._get_config()
        except ValidationError as exc:  # pragma: no cover
            sys.exit(str(exc))
        self.v_release()

    @classmethod
    def make_parser(
        cls,
        parser_func: cabc.Callable = argparse.ArgumentParser,
        *,
        add_help: bool = False,
        **kwargs,
    ) -> argparse.ArgumentParser:
        kwargs["description"] = cls.__doc__
        kwargs["parents"] = [cls.branch_repo_parser()]
        if add_help:
            kwargs["help"] = cls.__doc__

        parser: argparse.ArgumentParser = parser_func(**kwargs)
        parser.add_argument("--debug", action="store_true")
        parser.add_argument("--all", action="store_true")
        return parser

    def run(self) -> None:
        bm = self.backend.BaseMaker()
        self.release.make_base(self.config, base_maker=bm, fill_sack=False)
        self._enable_disable_bm(bm)
        for repo in bm.repolist(True):
            print(repo)



#!/usr/bin/env python3

# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import sys


def main(frag: str, file: str):
    with open(frag) as fp:
        frag_lines = tuple(fp)
    lines: list[str] = []
    with open(file, "r+") as fp:
        needs_append = True
        for line in fp:
            if needs_append and line.startswith("##"):
                if frag_lines[0] != line:
                    lines.extend((*frag_lines, "\n"))
                needs_append = False
            lines.append(line)
        fp.seek(0)
        fp.writelines(lines)
        fp.truncate()


if __name__ == "__main__":
    main(sys.argv[1], sys.argv[2])



#!/usr/bin/env python3

# Copyright (C) 2023 Maxwell G <maxwell@gtmx.me>
#
# SPDX-License-Identifier: GPL-2.0-or-later OR MIT

"""
Sync the upstream specfile with downstream.
Include gpg signature and verify it.
"""

import argparse
import subprocess
import specfile
from pathlib import Path


def parseargs() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("out", type=Path, nargs=1)
    return parser.parse_args()


def main():
    args = parseargs()
    out = args.out[0]
    with specfile.Specfile(out) as spec:
        with spec.sources() as sources:
            sources.insert_numbered(1, sources[0].location.strip() + ".asc")
            sources.insert_numbered(2, "https://meta.sr.ht/~gotmax23.pgp")
        with spec.sections() as sections:
            prep = sections.prep
            prep.insert(0, "%gpgverify -d0 -s1 -k2")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import logging

fmt = "{levelname}:{name}:{lineno}: {message}"
logging.basicConfig(format=fmt, style="{")
logger = logging.getLogger("fedrq")



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest

import fedrq.cli


@pytest.fixture
def run_command(capsys, patch_config_dirs):
    def runner(args):
        fedrq.cli.main(["subpkgs", "--sc", *args])
        stdout, stderr = capsys.readouterr()
        result = stdout.splitlines(), stderr.splitlines()
        return result

    return runner


def test_subpkgs_basic(run_command):
    out = run_command(["packagea", "-F", "name"])
    assert out[0] == ["packagea", "packagea-sub"]
    assert not out[1]


def test_subpkgs_specific_version(run_command, target_cpu):
    out = run_command(["packageb-11111:2-1.fc36.src"])
    expected = [
        f"packageb-11111:2-1.fc36.{target_cpu}",
        "packageb-sub-11111:2-1.fc36.noarch",
    ]

    assert out[0] == expected
    assert not out[1]


def test_subpkg_latest(run_command):
    """
    --latest=1 is the default, but let's test it anyways
    """
    out = run_command(["packageb", "--latest", "1", "-F", "nv"])
    expected1 = [
        "packageb-2",
        "packageb-sub-2",
    ]
    assert out[0] == expected1
    assert not out[1]


@pytest.mark.parametrize(
    "largs",
    (
        ["--latest=a"],
        ["--latest=all"],
    ),
)
def test_subpkg_all(largs: list[str], run_command, target_cpu):
    expected2 = [
        f"packageb-1-1.fc36.{target_cpu}",
        f"packageb-11111:2-1.fc36.{target_cpu}",
        "packageb-sub-1-1.fc36.noarch",
        "packageb-sub-11111:2-1.fc36.noarch",
    ]
    out2 = run_command([*largs, "packageb"])
    assert out2[0] == expected2
    assert not out2[1]


def test_subpkg_noarch(run_command):
    out = run_command(["packagea.src", "packageb.src", "-l=a", "--arch=noarch"])
    expected = [
        "packagea-1-1.fc36.noarch",
        "packagea-sub-1-1.fc36.noarch",
        "packageb-sub-1-1.fc36.noarch",
        "packageb-sub-11111:2-1.fc36.noarch",
    ]
    assert out[0] == expected
    assert not out[1]


def test_subpkg_arched(run_command, target_cpu):
    out = run_command(["packagea", "packageb", "-A", target_cpu, "-l", "ALL"])
    expected = [
        f"packageb-1-1.fc36.{target_cpu}",
        f"packageb-11111:2-1.fc36.{target_cpu}",
    ]
    assert out[0] == expected
    assert not out[1]


def test_subpkg_match(run_command):
    stdout, stderr = run_command(
        [
            "packagea-1",
            "packageb-1",
            "--match",
            "packagea*",
            "-M",
            "*-sub",
            "-Fna",
            "-la",
        ]
    )
    assert stdout == ["packagea.noarch", "packagea-sub.noarch", "packageb-sub.noarch"]
    assert not stderr



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

import json

import pytest

import fedrq.cli.formatters as formatters


def formatter(query, formatter_name="plain", *args, attr=False, **kwargs):
    result = sorted(
        (
            str(i)
            for i in formatters.DefaultFormatters.get_formatter(formatter_name).format(
                query, *args, **kwargs
            )
        )
    )
    if attr:
        assert result == sorted(
            (
                str(i)
                for i in formatters.DefaultFormatters.get_formatter(
                    f"attr:{formatter_name}"
                ).format(query, *args, **kwargs)
            )
        )
    return result


# @pytest.mark.parametrize("special_repos", ("repo1",), indirect=["special_repos"])
def test_plain_formatter(repo_test_rq, target_cpu):
    expected = sorted(
        (
            "packagea-1-1.fc36.noarch",
            "packagea-1-1.fc36.src",
            "packagea-sub-1-1.fc36.noarch",
            "packageb-1-1.fc36.src",
            f"packageb-1-1.fc36.{target_cpu}",
            "packageb-11111:2-1.fc36.src",
            f"packageb-11111:2-1.fc36.{target_cpu}",
            "packageb-sub-1-1.fc36.noarch",
            "packageb-sub-11111:2-1.fc36.noarch",
        )
    )
    query = repo_test_rq.query()
    assert formatter(query) == expected
    assert formatter(query, "plain") == expected


def test_plainwithrepo_formatter(repo_test_rq, target_cpu):
    expected = sorted(
        (
            "packagea-1-1.fc36.noarch testrepo1",
            "packagea-1-1.fc36.src testrepo1",
            "packagea-sub-1-1.fc36.noarch testrepo1",
            "packageb-1-1.fc36.src testrepo1",
            f"packageb-1-1.fc36.{target_cpu} testrepo1",
            "packageb-11111:2-1.fc36.src testrepo1",
            f"packageb-11111:2-1.fc36.{target_cpu} testrepo1",
            "packageb-sub-1-1.fc36.noarch testrepo1",
            "packageb-sub-11111:2-1.fc36.noarch testrepo1",
        )
    )
    query = repo_test_rq.query()
    assert formatter(query, "plainwithrepo") == expected
    assert formatter(query, "nevrr") == expected


def test_name_formatter(repo_test_rq):
    expected = sorted(
        (
            "packagea",
            "packagea",
            "packagea-sub",
            "packageb",
            "packageb",
            "packageb",
            "packageb",
            "packageb-sub",
            "packageb-sub",
        )
    )
    query = repo_test_rq.query()
    assert formatter(query, "name") == expected
    assert formatter(query, "attr:name") == expected


def test_evr_formatter(repo_test_rq):
    query = repo_test_rq.query(name__glob="packageb*")
    result = sorted(
        (
            "11111:2-1.fc36",
            "11111:2-1.fc36",
            "11111:2-1.fc36",
            "1-1.fc36",
            "1-1.fc36",
            "1-1.fc36",
        )
    )
    assert formatter(query, "evr") == result
    assert formatter(query, "attr:evr") == result


def test_nv_formatter(repo_test_rq):
    query = repo_test_rq.query(name__glob="packagea*")
    expected = sorted(("packagea-1", "packagea-1", "packagea-sub-1"))
    assert formatter(query, "nv") == expected


def test_source_formatter(repo_test_rq):
    query = repo_test_rq.query()
    assert formatter(query, "source") == ["packagea", "packageb"]


@pytest.mark.parametrize(
    "latest,expected",
    (
        (None, ["1", "1", "2", "2"]),
        (1, ["2", "2"]),
    ),
)
def test_version_formatter(repo_test_rq, latest, expected):
    query = repo_test_rq.query(name="packageb", latest=latest)
    assert formatter(query, "version") == expected
    assert formatter(query, "attr:version") == expected


def test_epoch_formatter(repo_test_rq):
    query = repo_test_rq.query(name="packageb-sub")
    assert formatter(query, "epoch") == ["0", "11111"]
    assert formatter(query, "attr:epoch") == ["0", "11111"]


def test_requires_formatter(repo_test_rq):
    query = repo_test_rq.query(name=("packagea-sub", "packageb-sub"))
    assert len(query) == 3
    expected = [
        "/usr/share/packageb-sub",
        "package(b)",
        "packagea = 1-1.fc36",
        "vpackage(b) = 1-1.fc36",
    ]
    assert formatter(query, "requires", attr=True) == expected


def test_repo_formatter(repo_test_rq):
    query = repo_test_rq.query()
    result = formatter(query, "reponame", attr=True)
    assert len(query) == len(result)
    assert {"testrepo1"} == set(result)


def test_repo_license_formatter(repo_test_rq):
    query = repo_test_rq.query(name__glob="packagea*")
    result = formatter(query, "license", attr=True)
    assert result == ["Unlicense"] * 3


def test_debug_name_formatter(repo_test_rq):
    query = repo_test_rq.query(name="packageb")
    result = formatter(query, "debug_name", attr=True)
    assert result == ["packageb-debuginfo"] * len(query)


def test_repo_files_formatter(repo_test_rq):
    query = repo_test_rq.query(name=["packagea", "packageb"], arch="notsrc", latest=1)
    result = formatter(query, "files", attr=True)
    assert result == ["/usr/share/packagea", "/usr/share/packageb"]


@pytest.mark.parametrize("attr", formatters._ATTRS)
def test_formatter_sanity(repo_test_rq, attr):
    """
    Sanity test to ensure that supported formatters work at all
    """
    query = repo_test_rq.query(
        name=["packagea", "packagea-sub", "packageb", "packageb-sub"], latest=1
    )
    result = formatter(query, attr)
    if attr not in (
        "provides",
        "requires",
        "obsoletes",
        "conflicts",
        "recommends",
        "suggests",
        "enhances",
        "supplements",
    ):
        assert len(result) == len(query)


def test_json_formatter(repo_test_rq):
    expected = [
        {
            "name": "packagea",
            "evr": "1-1.fc36",
            "arch": "noarch",
            "requires": ["vpackage(b)"],
            "conflicts": [],
            "provides": ["package(a)", "packagea = 1-1.fc36", "vpackage(a) = 1-1.fc36"],
            "source_name": "packagea",
        },
        {
            "name": "packagea",
            "evr": "1-1.fc36",
            "arch": "src",
            "requires": ["vpackage(b) > 0"],
            "conflicts": [],
            "provides": ["packagea = 1-1.fc36", "packagea-sub = 1-1.fc36"],
            "source_name": None,
        },
        {
            "name": "packagea-sub",
            "evr": "1-1.fc36",
            "arch": "noarch",
            "requires": ["/usr/share/packageb-sub", "packagea = 1-1.fc36"],
            "conflicts": [],
            "provides": [
                "subpackage(a)",
                "packagea-sub = 1-1.fc36",
                "vsubpackage(a) = 1-1.fc36",
            ],
            "source_name": "packagea",
        },
    ]
    query = repo_test_rq.resolve_pkg_specs(["packagea*"], latest=1)
    output = formatter(
        query, "json:name,evr,arch,requires,conflicts,provides,source_name"
    )
    assert len(output) == 1
    assert json.loads(output[0]) == expected



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
fedrq supports dnf and libdnf5. Importing both in the same process doesn't
work, so we use some finagling to dynmaically load the requested backend.
"""

from __future__ import annotations

import functools
import logging
import warnings
from typing import TYPE_CHECKING

from fedrq.backends import dnf, libdnf5

if TYPE_CHECKING:
    from types import ModuleType

    from .base import BackendMod


class MissingBackendError(Exception):
    pass


BACKENDS: dict[str, ModuleType] = {
    "dnf": dnf,
    "libdnf5": libdnf5,
}
DEFAULT_BACKEND = "dnf"
LOG = logging.getLogger(__name__)


class Backends:
    def __init__(self) -> None:
        self.available: dict[str, ModuleType] = {}
        self.missing: dict[str, MissingBackendError] = {}
        for name, mod in BACKENDS.items():
            try:
                mod.ensure_backend()
            except MissingBackendError as exc:
                self.missing[name] = exc
            else:
                self.available[name] = mod


@functools.lru_cache
def _get_backends() -> Backends:
    return Backends()


class _DefaultBackend:
    def __init__(self) -> None:
        self.backend: BackendMod | None = None

    def __call__(
        self, default: str | None = None, fallback: bool = False
    ) -> BackendMod:
        if not self.backend:
            self.backend = self._get_backend(default, fallback)
            LOG.info("Using %s backend", self.backend.BACKEND)
        elif default and default != self.backend.BACKEND:
            warnings.warn(
                f"Falling back to {self.backend.BACKEND}. {default} cannot be used."
            )
        return self.backend

    def _get_backend(
        self, default: str | None = None, fallback: bool = False
    ) -> BackendMod:
        default = default or DEFAULT_BACKEND
        if default not in BACKENDS:
            raise MissingBackendError(f"Invalid backend {default!r}.")
        backends = _get_backends()
        if backend := backends.available.get(default):
            return backend.get_backend()
        if not fallback:
            raise backends.missing[default]
        if backend := next(iter(backends.available.values()), None):
            return backend.get_backend()
        _split = ", ".join(backends.missing)
        raise MissingBackendError(
            f"Neither of the following backends were available: {_split}"
        )


get_default_backend = _DefaultBackend()



# Copyright (C) 2023 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest


@pytest.mark.no_rpm_mock
@pytest.mark.parametrize(
    "args",
    [
        pytest.param(["ansible-core"]),
        pytest.param(["ansible-core.src"]),
    ],
)
def test_whatrequires_src_integration(runs, args):
    stdout, stderr = runs(["wrsrc", "-b", "f37", "-Fna", *args], False)
    assert not stderr
    assert any(p.startswith("ansible-collection") for p in stdout)
    assert {"ansible-packaging.noarch", "ansible.noarch", "ansible.src"} & set(stdout)


@pytest.mark.no_rpm_mock
def test_whatrequires_src_integration_exclude(runs):
    stdout, stderr = runs(
        ["wrsrc", "-b", "f37", "-Fsource", "-X", "ansible-packaging"], False
    )
    assert not stderr
    assert {
        "ansible-collection-ansible-posix",
        "ansible-collection-community-general",
    } & set(stdout)
    assert "ansible-packaging" not in stdout


@pytest.mark.no_rpm_mock
def test_whatrequires_src_integration_exclude_control(runs):
    stdout, stderr = runs(
        ["wrsrc", "-b", "f37", "-Fsource", "ansible-packaging"], False
    )
    assert not stderr
    assert {
        "ansible-collection-ansible-posix",
        "ansible-collection-community-general",
    } & set(stdout)
    assert "ansible-packaging" in stdout



#!/usr/bin/env python3
# SPDX-License-Identifier: Unlicense
# SPDX-FileCopyrightText: None

"""
This is a commented, fully typed example script showing fedrq's API.
It searches the Fedora repositories for packages that haven't been rebuilt
in the last few releases (i.e. Long Term FTBFS packages).
Based on https://github.com/hroncok/fedora-report-ftbfs-retirements.
"""
from __future__ import annotations

import argparse

from fedrq.backends.base import PackageCompat, PackageQueryCompat, RepoqueryBase
from fedrq.config import RQConfig, get_config

DEFAULT_RELEASES = ".fc36", ".fc37", ".fc38", ".fc39"
DEFAULT_IGNORES = (
    "dummy-test-package",
    "fedora-obsolete-packages",
    "fedora-release",
    "fedora-repos",
    "generic-release",
    "shim",
)


def parseargs() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-R",
        "--releases",
        nargs="*",
        help=f"DEFAULT: {DEFAULT_RELEASES}",
        default=DEFAULT_RELEASES,
    )
    args = parser.parse_args()
    return args


def main(*, releases: list[str]):
    # get_config() loads the configuration from the filesystem as explained in
    # fedrq(5).
    # Extra kwargs may be passed to get_config() to override certain values.
    # For example, you can pass `backend="libdnf5"` if you'd like to change the
    # default backend.
    config: RQConfig = get_config(backend="libdnf5")

    # Query the Fedora Rawhide mirrored repositories.
    # This supports any release configuration builtin to fedrq
    # or configured on your local system.
    rq: RepoqueryBase = config.get_rq("rawhide")

    # PackageQueryCompat is a Protocol implemented by both backends.
    # It is an iterable of PackageCompat objects with extra methods to
    # filter packages and preform other similar tasks.
    #
    # rq.query() creates a PackageQuery instance and calls its filterm() method
    # with the provided kwargs.

    # Create a Query with only source packages
    query: PackageQueryCompat = rq.query(arch="src")

    # PackageCompat is a Protocol implemented by both the dnf and libdnf5
    # backends.
    packages: list[PackageCompat] = []

    for package in query:
        if (
            not any(package.name.startswith(i) for i in DEFAULT_IGNORES) and
            not any(i in package.release for i in releases)
        ):
            packages.append(package)
    for p in packages:
        print(p.name)


if __name__ == "__main__":
    main(**vars(parseargs()))



# SPDX-FileCopyrightText: 2023 Maxwell G <gotmax@e.email>
#
# SPDX-License-Identifier: GPL-2.0-or-later OR MIT

from __future__ import annotations

import itertools
import os
import subprocess
from collections.abc import Collection, Iterable
from pathlib import Path
from shutil import copy2, rmtree

import nox
import nox.command
import nox.virtualenv

IN_CI = "JOB_ID" in os.environ
ALLOW_EDITABLE = os.environ.get("ALLOW_EDITABLE", str(not IN_CI)).lower() in (
    "1",
    "true",
)
LINT_SESSIONS = ("formatters", "codeql", "typing", "reuse")
PROJECT = "fedrq"

nox.options.sessions = (*LINT_SESSIONS, "test", "libdnf5_test")


def install(session: nox.Session, *args, editable=False, **kwargs):
    if isinstance(session.virtualenv, nox.virtualenv.PassthroughEnv):
        session.warn(f"No venv. Skipping installation of {args}")
        return
    if editable and ALLOW_EDITABLE:
        args = ("-e", *args)
    session.install(*args, "-U", **kwargs)


def run_silent(*args, return_stdout: bool = False, **kwargs):
    kwargs.setdefault("text", True)
    kwargs.setdefault("stdout", subprocess.PIPE)
    kwargs.setdefault("check", True)
    proc = subprocess.run(args, **kwargs)
    return proc.stdout if return_stdout else proc


def git(session: nox.Session, *args, **kwargs):
    return session.run("git", *args, **kwargs, external=True)


def _to_install_system(session, *packages: str):
    for package in packages:
        for whatprovides in [(), ("--whatprovides",), "yield"]:
            if whatprovides == "yield":
                session.log(f"Installing RPM package {package!r}")
                yield package
            if not subprocess.run(
                ["rpm", "-q", *whatprovides, package],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            ).returncode:
                session.log(f"RPM package {package!r} is already installed.")
                break


def install_system(
    session: nox.Session, *packages: str, install_weak_deps: bool = False
):
    to_install = list(_to_install_system(session, *packages))
    if not to_install:
        return
    cmd = ["dnf", "install"]
    if not session.interactive:
        cmd.append("-y")
    if os.geteuid() != 0:
        cmd.insert(0, "sudo")
    if not install_weak_deps:
        cmd.append("--setopt=install_weak_deps=False")
    session.run_always(*cmd, *to_install, external=True)


@nox.session(venv_params=["--system-site-packages"])
def test(session: nox.Session, backend=None):
    install_system(session, "createrepo_c", "rpm-build", "python3-rpm")
    install(session, ".[test]", "pytest-xdist", editable=True)
    posargs = session.posargs
    if "--check" in posargs:
        posargs.remove("--check")
    session.run(
        "python",
        "-m",
        "pytest",
        *session.posargs,
        env={
            "PYTEST_PLUGINS": "xdist.plugin,pytest_mock",
            "PYTEST_DISABLE_PLUGIN_AUTOLOAD": "1",
        }
        | ({"FEDRQ_BACKEND": backend} if backend else {}),
    )


@nox.session(venv_backend="none")
def testa(session):
    session.notify("test")
    session.notify("libdnf5_test")


@nox.session(venv_params=["--system-site-packages"])
def libdnf5_test(session: nox.Session):
    install_system(session, "python3-libdnf5")
    test(session, "libdnf5")


@nox.session(venv_backend="none")
def lint(session: nox.Session):
    """
    Run format, codeql, typing, and reuse sessions
    """
    for notify in LINT_SESSIONS:
        session.notify(notify)


@nox.session()
def codeql(session: nox.Session):
    install(
        session,
        "ruff",
    )
    posargs = session.posargs
    if "--check" in posargs:
        posargs.remove("--check")
    session.run(
        "ruff",
        *session.posargs,
        "src/fedrq/",
        "tests/",
        "noxfile.py",
    )


@nox.session
def typing(session: nox.Session):
    install(session, ".[lint]", "-I", editable=True)
    session.run(
        "python", "-m", "mypy", "--enable-incomplete-feature=Unpack", "src/fedrq/"
    )


@nox.session
def formatters(session: nox.Session):
    install(session, "black", "isort")
    posargs = session.posargs
    if IN_CI:
        posargs = ["--check"]
    try:
        session.run(
            "python",
            "-m",
            "black",
            *posargs,
            "src/fedrq/",
            "tests/",
            "noxfile.py",
        )
    finally:
        try:
            session.run(
                "python",
                "-m",
                "isort",
                "--add-import",
                "from __future__ import annotations",
                *posargs,
                "src/fedrq/",
                "noxfile.py",
            )
        finally:
            session.run("python", "-m", "isort", *posargs, "tests/")


@nox.session
def reuse(session: nox.Session):
    install(session, "reuse")
    session.run("reuse", "lint")


def install_fclogr(session: nox.Session, allow_local: bool = True):
    install_system(session, "rpm-build", "python3-rpm")
    if allow_local and Path("../fclogr").exists():
        install(session, "-e", "../fclogr")
    else:
        install(session, "git+https://git.sr.ht/~gotmax23/fclogr#main")


def _spec_changed(session: nox.Session) -> bool:
    if "fedrq.spec" not in run_silent("git", "diff", "--name-only", return_stdout=True):
        return False
    if not session.interactive:
        return True
    session.run("git", "diff", "fedrq.spec", external=True, env=dict(PAGER=""))
    session.log(
        "fedrq.spec has changed in the working tree. Should we revert it? (y/N)"
    )
    return input().lower() == "y"


@nox.session(venv_backend="none")
def clean(session: nox.Session):
    exts = ("fedrq-*.tar.gz", "fedrq-*.src.rpm")
    for file in itertools.chain.from_iterable(Path().glob(ext) for ext in exts):
        session.log(f"Removing old artifact: {file}")
        file.unlink()
    if _spec_changed(session):
        session.run("git", "restore", "fedrq.spec")


@nox.session
def srpm(session: nox.Session, posargs=None):
    posargs = posargs or session.posargs
    install_fclogr(session)
    session.run("fclogr", "--debug", "dev-srpm", *posargs)


@nox.session
def mockbuild(session: nox.Session):
    install_system(session, "mock", "mock-core-configs")
    tmp = Path(session.create_tmp())
    srpm(session, ("-o", tmp, "--keep"))
    spec_path = tmp / "fedrq.spec"
    margs = [
        "mock",
        "--spec",
        str(spec_path),
        "--source",
        str(tmp),
        *session.posargs,
    ]
    if not session.interactive:
        margs.append("--verbose")
    if not {
        "--clean",
        "--no-clean",
        "--cleanup-after",
        "--no-cleanup-after",
        "-n",
        "-N",
    } & set(session.posargs):
        margs.insert(1, "-N")
    session.run(*margs, external=True)


def ensure_clean(session: nox.Session):
    if git(session, "status", "--porcelain", "--untracked-files", silent=True):
        msg = "There are untracked and/or modified files."
        session.error(msg)


def _check_git_tag(session: nox.Session, version: str):
    tag = "v" + version
    tags = git(session, "tag", "--list", silent=True).splitlines()
    if tag in tags:
        session.error(f"{tag} is already tagged")


@nox.session
def bump(session: nox.Session):
    version = session.posargs[0]
    _check_git_tag(session, version)
    ensure_clean(session)

    install(session, "flit>=3.9", "tomcli[tomlkit]", "twine")
    install_fclogr(session, False)

    # Download generate_changelog
    tmp = Path(session.create_tmp())
    with session.chdir(tmp):
        session.run_always(
            "wget",
            "https://git.sr.ht/~gotmax23/fedora-scripts/blob/main/generate_changelog.py",
            external=True,
        )

    # Update version
    session.run("tomcli-set", "pyproject.toml", "str", "project.version", version)

    install(session, ".")

    # Bump specfile
    session.run(
        "fclogr",
        "bump",
        "--new",
        version,
        "--comment",
        f"Release {version}.",
        f"{PROJECT}.spec",
    )

    # Bump changelog, commit, and tag
    git(session, "add", f"{PROJECT}.spec", "pyproject.toml")
    session.run(
        "python", str(tmp / "generate_changelog.py"), f"--version={version}", "--tag"
    )

    if Path("dist").is_dir():
        rmtree("dist")
    session.run("flit", "build", "--use-vcs")
    artifacts = _get_artifacts(("*.whl", "*.tar.gz"), expected_count=2)
    session.run("twine", "check", "--strict", *artifacts)
    _sign_artifacts(session)


def _sign_artifacts(session: nox.Session) -> None:
    uid = git(session, "config", "user.email", silent=True).strip()
    artifacts = _get_artifacts(["tar.gz", "whl"], expected_count=2)
    for path in artifacts:
        if Path(path + ".asc").exists():
            session.warn(f"{path}.asc already exists. Not signing it.")
            continue
        session.run(
            "gpg", "--local-user", uid, "--armor", "--detach-sign", path, external=True
        )


def _get_artifacts(
    exts: Collection[str] = ("whl", "tar.gz", "tar.gz.asc", "whl.asc"),
    required=True,
    extra_allowed=True,
    expected_count: int | None = None,
) -> Iterable[str]:
    count = 0
    all_files = set(Path("dist").iterdir())
    for ext in exts:
        found = tuple(Path("dist").glob(f"*.{ext}"))
        if not found and required:
            raise ValueError(f"No matches for {ext}")
        count += len(found)
        for file in found:
            all_files.remove(file)
        yield from map(str, found)
    if all_files and not extra_allowed:
        raise ValueError(f"Extra files found in dist: {all_files}")
    if expected_count is not None and count != expected_count:
        raise ValueError(f"Found {count} artifacts. Expected {expected_count}.")


@nox.session
def publish(session: nox.Session):
    # Setup
    ensure_clean(session)
    install(session, "twine", "tomcli[tomlkit]")
    session.run(
        "twine",
        "upload",
        "--non-interactive",
        "--username",
        "__token__",
        *_get_artifacts(("*.whl", "*.tar.gz"), extra_allowed=False, expected_count=2),
    )

    # Copr build
    copr_release(session)

    # Push to git
    if session.interactive and input("Push to Sourcehut (Y/n)").lower() != "n":
        git(session, "push", "--follow-tags")
        srht_artifacts(session)

    # Post-release bump
    version = session.run(  # type: ignore[union-attr]
        "tomcli-get", "pyproject.toml", "project.version", silent=True
    ).strip()
    session.run(
        "tomcli-set", "pyproject.toml", "str", "project.version", f"{version}.post0"
    )
    git(session, "add", "pyproject.toml")
    git(session, "commit", "-S", "-m", "Post release version bump")


@nox.session
def docgen(session: nox.Session):
    """
    Generate extra content for the docsite
    """
    for i in ("1", "5"):
        # Long, terrible pipeline to convert scdoc to markdown
        # fmt: off
        session.run(
            "sh", "-euo", "pipefail", "-c",
            # Convert scdoc to html
            f"scd2html < doc/fedrq.{i}.scd"
            # Remove aria-hidden attributes so pandoc doesn't try to convert them
            "| sed 's|aria-hidden=\"true\"||'"
            "| pandoc --from html "
            # mkdocs doesn't support most of the pandoc markdown extensions.
            # Use markdown_strict and only enable pipe_tables.
            "--to markdown_strict+pipe_tables"
            "| sed "
            # Remove anchors that scd2html inserts
            r"-e 's| \[¶\].*||' "
            f"> doc/fedrq{i}.md",
            external=True,
        )
        # fmt: on


@nox.session
def copr_release(session: nox.Session):
    install(session, "copr-cli", "requests-gssapi", "specfile")
    tmp = Path(session.create_tmp())
    dest = tmp / "tomcli.spec"
    copy2("tomcli.spec", dest)
    session.run("python", "contrib/fedoraify.py", str(dest))
    session.run("copr-cli", "build", "--nowait", "gotmax23/tomcli", str(dest))


@nox.session(python="none")
def srht_artifacts(session: nox.Session):
    artifacts = map(str, Path("dist").glob("*"))
    session.run("hut", "git", "artifact", "upload", *artifacts, external=True)


@nox.session
def mkdocs(session: nox.Session):
    install(session, "-e", ".[doc]")
    docgen(session)
    session.run("mkdocs", *session.posargs)



# SPDX-FileCopyrightText: 2022 Maxwell G <gotmax@e.email>
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest

import fedrq.cli


@pytest.fixture
def run_command(capsys, patch_config_dirs):
    def runner(args, *, return_tuples=False):
        fedrq.cli.main(["whatrequires", "--sc", *args])
        stdout, stderr = capsys.readouterr()
        result = stdout.splitlines(), stderr.splitlines()
        if return_tuples:
            result = tuple(tuple(r) for r in result)
        return result

    return runner


def test_whatrequires_exact(run_command):
    output = run_command(["package(b)"])
    output2 = run_command(["package(b)", "-E"])
    assert output == output2
    assert output[0] == ["packageb-sub-11111:2-1.fc36.noarch"]
    assert not output[1]


def test_whatrequires_name(run_command):
    output = run_command(["packageb", "-l", "a", "-F", "name"])
    assert output[0] == ["packagea"] * 2 + ["packageb-sub"] * 2
    assert not output[1]


@pytest.mark.parametrize(
    "args",
    (
        pytest.param(["package(b)", "-l", "a", "-P"]),
        pytest.param(["vpackage(b)", "-l", "a", "-P"]),
        pytest.param(["packageb", "-l", "a", "-P"]),
        pytest.param(
            ["packageb", "-l", "a"],
        ),
        pytest.param(["/usr/share/packageb", "-l", "a", "-P", "-Lalways"]),
    ),
)
def test_whatrequires_resolve_b(run_command, args):
    stdout, stderr = run_command(args, return_tuples=True)
    assert stdout == (
        "packagea-1-1.fc36.noarch",
        "packagea-1-1.fc36.src",
        "packageb-sub-1-1.fc36.noarch",
        "packageb-sub-11111:2-1.fc36.noarch",
    )
    assert not stderr


@pytest.mark.parametrize(
    "args",
    (
        (["packagea", "-E"]),
        (["packagea", "-P"]),
        (["package(a)", "-P"]),
        (["vpackage(a)", "-P"]),
        (["/usr/share/packagea", "-P", "-Lalways"]),
    ),
)
def test_whatrequires_resolve_a(run_command, args):
    output = run_command(args + ["-F", "nv"])
    assert output[0] == ["packagea-sub-1"]
    assert not output[1]


def test_whatrequires_versioned_resolve(run_command):
    output = run_command(
        [
            "vpackage(b) = 11111:2-1.fc36",
            "-P",
            "-l",
            "all",
        ]
    )
    assert output[0] == [
        "packagea-1-1.fc36.noarch",
        "packagea-1-1.fc36.src",
        "packageb-sub-11111:2-1.fc36.noarch",
    ]
    assert not output[1]


@pytest.mark.parametrize(
    "args, exact_optional",
    (
        # Choose a random formatter to check that they don't fail
        # when no packages are provided.
        (["package(a)", "-F", "attr:repoid"], True),
        (["vpackage(a)", "-F", "source"], True),
        (["/usr/share/packagea", "-F", "attr:sourcerpm", "-Lalways"], True),
        (["/usr/share/packageb", "-F", "nev", "-Lalways"], True),
        # fedrq will resolve package names, so we need
        # to explicitly pass -E.
        (["packageb", "-F", "nv"], False),
        (["packageb.{target_cpu}", "-F", "na"], False),
    ),
)
def test_exact_no_result(args, exact_optional, run_command, target_cpu):
    """
    These work with -P, but should not print any results
    with --exact.
    """
    expected = ([], [])
    args = [arg.format(target_cpu=target_cpu) for arg in args]
    output = run_command(args + ["-E"])
    assert output == expected
    if exact_optional:
        output2 = run_command(args)
        assert output2 == expected


def test_whatrequires_breakdown(run_command):
    expected = """\
Runtime:
packagea
packageb-sub
    2 total runtime dependencies

Buildtime:
packagea
    1 total buildtime dependencies

All SRPM names:
packagea
packageb
    2 total SRPMs""".splitlines()
    output = run_command(["-F", "breakdown", "packageb"])
    assert output[0] == expected
    assert not output[1]


def test_whatrequires_exclude_subpackages(run_command):
    expected = ["packagea.noarch", "packagea.src"]
    stdout, stderr = run_command(["-X", "-F", "na", "packageb"])
    assert not stderr
    assert stdout == expected
